import pandas as pd
from ..exceptions import ColumnValidationError
from typing import Optional

def validate_min(series: pd.Series, min_value: Optional[float], column_name: str):
    if min_value is None:
        return
    if series is None:
        raise ColumnValidationError(column_name, "column missing", sample=[])
    # Compare only on non-null numeric-friendly values
    try:
        mask = series < min_value
    except Exception:
        # If comparison fails due to types, show offending sample
        sample = series[~series.apply(lambda v: pd.isna(v) or isinstance(v, (int, float)))].head(10).tolist()
        raise ColumnValidationError(column_name, f"cannot compare values with min {min_value}", sample=sample)
    if mask.any():
        sample = series[mask].head(10).tolist()
        raise ColumnValidationError(column_name, f"values below min ({min_value})", sample=sample)

def validate_max(series: pd.Series, max_value: Optional[float], column_name: str):
    if max_value is None:
        return
    if series is None:
        raise ColumnValidationError(column_name, "column missing", sample=[])
    try:
        mask = series > max_value
    except Exception:
        sample = series[~series.apply(lambda v: pd.isna(v) or isinstance(v, (int, float)))].head(10).tolist()
        raise ColumnValidationError(column_name, f"cannot compare values with max {max_value}", sample=sample)
    if mask.any():
        sample = series[mask].head(10).tolist()
        raise ColumnValidationError(column_name, f"values above max ({max_value})", sample=sample)



import pandas as pd
from ..exceptions import ColumnValidationError

def validate_unique(series: pd.Series, column_name: str):
    """
    Ensure values in series are unique (ignores nulls).
    Raises ColumnValidationError with sample duplicate rows.
    """
    if series is None:
        raise ColumnValidationError(column_name, "column missing", sample=[])
    # Consider only non-null values for uniqueness
    non_null = series.dropna()
    dup_mask = non_null.duplicated(keep=False)
    if dup_mask.any():
        dup_values = non_null[dup_mask].head(10).tolist()
        raise ColumnValidationError(column_name, "duplicate values found", sample=dup_values)



import pandas as pd
from ..exceptions import ColumnValidationError
import re

def validate_regex(series: pd.Series, pattern: str, column_name: str):
    """
    Validate that string (non-null) values match regex `pattern`.
    Non-string values will be treated as mismatch (except nulls).
    """
    if pattern is None:
        return
    if series is None:
        raise ColumnValidationError(column_name, "column missing", sample=[])
    # safe vectorized check: convert to str but keep track of originally non-null non-str values
    # We want to treat non-string non-null as mismatch.
    non_null = series.dropna()
    if non_null.empty:
        return
    # Build mask of matches; safe with exceptions
    try:
        # Ensure pattern is valid
        re.compile(pattern)
        # use pandas str.match (works on nan-safe series if converted to str)
        matches = non_null.astype(str).str.match(pattern)
        mismatches = non_null[~matches]
        if not mismatches.empty:
            sample = mismatches.head(10).tolist()
            raise ColumnValidationError(column_name, f"regex mismatch ({pattern})", sample=sample)
    except re.error as e:
        raise ColumnValidationError(column_name, f"invalid regex pattern: {e}", sample=[])



import yaml
from ..column import Column
from ..schema import Schema
from typing import Dict, Any

# Map type names (strings in YAML) to python types
_TYPE_MAP = {
    "int": int,
    "integer": int,
    "float": float,
    "double": float,
    "str": str,
    "string": str,
    "bool": bool,
    "boolean": bool,
}


def _resolve_type(type_value):
    if isinstance(type_value, str):
        t = _TYPE_MAP.get(type_value.lower())
        if t is None:
            raise ValueError(f"Unknown type name in schema: {type_value}")
        return t
    return type_value  # assume it's already a python type

def load_schema_from_yaml(path: str) -> Schema:
    """
    Load schema YAML and return a Schema object.
    Example YAML:
    version: "1.0"
    columns:
      order_id:
        type: int
        not_null: true
        unique: true
      amount:
        type: float
        min: 0
        max: 10000
      email:
        type: string
        regex: ".+@.+\\.com"
    """
    with open(path, "r", encoding="utf-8") as fh:
        data = yaml.safe_load(fh)

    version = str(data.get("version", "0"))
    cols = []
    for name, cfg in data.get("columns", {}).items():
        dtype = _resolve_type(cfg.get("type", "string"))
        col = Column(
            name=name,
            dtype=dtype,
            not_null=bool(cfg.get("not_null", False)),
            unique=bool(cfg.get("unique", False)),
            min_value=cfg.get("min", None),
            max_value=cfg.get("max", None),
            regex=cfg.get("regex", None),
        )
        cols.append(col)

    return Schema(version=version, columns=cols)



from .schema import Schema
from .column import Column
from .loader.yaml_loader import load_schema_from_yaml
from .exceptions import ValidationError, ColumnValidationError, SchemaVersionError

__all__ = [
    "Schema",
    "Column",
    "load_schema_from_yaml",
    "ValidationError",
    "ColumnValidationError",
    "SchemaVersionError",
]



import pandas as pd
from typing import Type
from ..exceptions import ColumnValidationError

# Allowed simple type mapping checkers (can be extended)
_TYPE_CHECKERS = {
    int: lambda x: isinstance(x, (int,)) and not isinstance(x, bool),
    float: lambda x: isinstance(x, (float, int)) and not isinstance(x, bool),
    str: lambda x: isinstance(x, str),
    bool: lambda x: isinstance(x, bool),
}

def _is_instance(value, expected_type: Type):
    if pd.isna(value):
        return True  # treat NaN/None as pass for type check (null check handled separately)
    checker = _TYPE_CHECKERS.get(expected_type)
    if checker:
        return checker(value)
    # fallback: use isinstance
    try:
        return isinstance(value, expected_type)
    except Exception:
        return False

def validate_type(series: pd.Series, expected_type: Type, column_name: str):
    """
    Validate that non-null values in `series` are of `expected_type`.
    Raises ColumnValidationError on failure with sample offending values.
    """
    if series is None:
        raise ColumnValidationError(column_name, "column missing", sample=[])
    mask = series.apply(lambda v: not _is_instance(v, expected_type))
    if mask.any():
        sample = series[mask].dropna().head(10).tolist()
        raise ColumnValidationError(column_name, f"type mismatch (expected {expected_type.__name__})", sample=sample)



from dataclasses import dataclass
from typing import Optional, Any


@dataclass
class Column:
    name: str
    dtype: Any  # python type like int, float, str, bool
    not_null: bool = False
    unique: bool = False
    min_value: Optional[float] = None
    max_value: Optional[float] = None
    regex: Optional[str] = None

    def __repr__(self):
        return (
            f"Column(name={self.name!r}, dtype={self.dtype}, not_null={self.not_null}, "
            f"unique={self.unique}, min={self.min_value}, max={self.max_value}, regex={self.regex})"
        )



from .schema import Schema
from typing import Dict, Any

def compare_schemas(old: Schema, new: Schema) -> Dict[str, Any]:
    """
    Compare two Schema objects and return a dict describing:
      - added_columns
      - removed_columns
      - changed_columns: {col: {old: <props>, new: <props>, breaking: bool}}
      - breaking: overall bool (True if any breaking change)
    Rules for breaking changes:
      - removed column -> breaking
      - change of dtype -> breaking
      - change nullable True->False -> breaking
    Adding a new column (nullable) is non-breaking.
    """
    old_cols = set(old.columns.keys())
    new_cols = set(new.columns.keys())

    added = list(new_cols - old_cols)
    removed = list(old_cols - new_cols)
    changed = {}

    breaking = False

    for col in (old_cols & new_cols):
        o = old.columns[col]
        n = new.columns[col]
        col_changes = {}
        if o.dtype != n.dtype:
            col_changes["dtype"] = {"old": getattr(o.dtype, "__name__", str(o.dtype)),
                                    "new": getattr(n.dtype, "__name__", str(n.dtype))}
            col_changes["breaking"] = True
            breaking = True
        if o.not_null and not n.not_null:
            # making required -> optional is compatible
            col_changes.setdefault("nullable", {})["old"] = o.not_null
            col_changes["nullable"]["new"] = n.not_null
            col_changes["nullable"]["breaking"] = False
        if (not o.not_null) and n.not_null:
            # optional -> required = breaking
            col_changes.setdefault("nullable", {})["old"] = o.not_null
            col_changes["nullable"]["new"] = n.not_null
            col_changes["nullable"]["breaking"] = True
            breaking = True
        if o.unique != n.unique:
            col_changes["unique"] = {"old": o.unique, "new": n.unique, "breaking": False}
            # making unique may be breaking implicitly but flagging as non-breaking for now
        if o.min_value != n.min_value or o.max_value != n.max_value:
            col_changes["bounds"] = {"old_min": o.min_value, "new_min": n.min_value,
                                     "old_max": o.max_value, "new_max": n.max_value,
                                     "breaking": False}

        if col_changes:
            changed[col] = col_changes

    # Removing column is breaking
    if removed:
        breaking = True

    return {
        "added_columns": added,
        "removed_columns": removed,
        "changed_columns": changed,
        "breaking": breaking
    }



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    description = fh.read()

setup(
    name="schemaguard",
    version="0.2.0",
    author="Inzamam Yousaf",
    author_email="uniprecisionofficial@gmail.com",
    description="Lightweight Schema Validator",
    url="https://github.com/inzamam1121/schemaguard",
    packages=find_packages(exclude=["examples*", "tests*"]),
    python_requires=">=3.8",
    install_requires=[
        "pandas>=1.0",
        "pyyaml>=6.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    long_description = description,
    long_description_content_type = "text/markdown",
)



from .type_validator import validate_type
from .null_validator import validate_not_null
from .unique_validator import validate_unique
from .bounds_validator import validate_min, validate_max
from .regex_validator import validate_regex

__all__ = [
    "validate_type",
    "validate_not_null",
    "validate_unique",
    "validate_min",
    "validate_max",
    "validate_regex",
]



class ValidationError(Exception):
    """Raised when schema validation fails for the dataframe as a whole."""

class ColumnValidationError(ValidationError):
    """
    Raised when a particular column fails validation.
    Attributes:
        column: name of the column
        reason: short reason string
        sample: a small sample of offending values (list)
    """
    def __init__(self, column: str, reason: str, sample=None):
        self.column = column
        self.reason = reason
        self.sample = sample or []
        message = f"Validation failed for column '{column}': {reason}."
        if self.sample:
            message += f" Sample offending values: {self.sample}"
        super().__init__(message)

class SchemaVersionError(Exception):
    """Errors related to schema versioning / compatibility."""



from typing import Dict, List, Optional
from .column import Column
import pandas as pd
from .exceptions import ColumnValidationError, ValidationError
from .validators import (
    validate_type,
    validate_not_null,
    validate_unique,
    validate_min,
    validate_max,
    validate_regex,
)

class Schema:
    def __init__(self, version: str, columns: Optional[List[Column]] = None):
        self.version = version
        self.columns: Dict[str, Column] = {}
        for c in (columns or []):
            self.columns[c.name] = c

    def add_column(self, column: Column):
        self.columns[column.name] = column

    def get_column(self, name: str) -> Optional[Column]:
        return self.columns.get(name)

    def validate(self, df: pd.DataFrame, raise_on_error: bool = True) -> Dict[str, Dict]:
        """
        Validate DataFrame against the schema.
        Returns a report dict:
            { "ok": bool, "errors": { column_name: { "reason": str, "sample": [...] } } }
        If raise_on_error is True and any error occurs, raises ValidationError with message and details.
        """
        errors = {}
        # ensure df is a DataFrame
        if not isinstance(df, pd.DataFrame):
            raise ValidationError("Input is not a pandas DataFrame.")

        for col_name, col in self.columns.items():
            # get series (None if missing)
            series = df[col_name] if col_name in df.columns else None
            try:
                # type check first
                validate_type(series, col.dtype, col_name)
                # null check
                if col.not_null:
                    validate_not_null(series, col_name)
                # unique
                if col.unique:
                    validate_unique(series, col_name)
                # bounds
                if col.min_value is not None:
                    validate_min(series, col.min_value, col_name)
                if col.max_value is not None:
                    validate_max(series, col.max_value, col_name)
                # regex
                if col.regex is not None:
                    validate_regex(series, col.regex, col_name)
            except ColumnValidationError as e:
                errors[col_name] = {"reason": e.reason, "sample": e.sample}

        ok = len(errors) == 0
        report = {"ok": ok, "errors": errors, "version": self.version}
        if not ok and raise_on_error:
            # Build explanatory message
            parts = []
            for cn, info in errors.items():
                parts.append(f"{cn}: {info['reason']} (sample: {info.get('sample', [])})")
            raise ValidationError("Schema validation failed: " + "; ".join(parts))
        return report



import pandas as pd
from ..exceptions import ColumnValidationError

def validate_not_null(series: pd.Series, column_name: str):
    """Raise if any value is null/NaN in series."""
    if series is None:
        raise ColumnValidationError(column_name, "column missing", sample=[])
    null_mask = series.isnull()
    if null_mask.any():
        sample = series[null_mask].head(10).tolist()
        raise ColumnValidationError(column_name, "null values present", sample=sample)
